# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "0.5.0"

# app name to send as part of SDK requests
app_name = "DLHub SDK v{}".format(__version__)
